"""
    lock_funds_transaction
    ======================

    Lock-funds transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .signed_transaction import SignedTransaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic import Mosaic
from ..mosaic.mosaic_id import MosaicId
from ..namespace.namespace_id import NamespaceId
from ... import util

__all__ = [
    'LockFundsTransaction',
    'LockFundsInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('LOCK')
class LockFundsTransaction(Transaction):
    """
    Lock-funds transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param mosaic: Locked mosaic.
    :param duration: Duration to lock funds.
    :param signed_transaction: Signed transaction for which funds are locked.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    mosaic: Mosaic
    duration: int
    signed_transaction: SignedTransaction

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        mosaic: Mosaic,
        duration: int,
        signed_transaction: SignedTransaction,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        if signed_transaction.type != TransactionType.AGGREGATE_BONDED:
            raise ValueError('Signed transaction must be aggregated bonded transaction.')
        super().__init__(
            TransactionType.LOCK,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('mosaic', mosaic)
        self._set('duration', duration)
        self._set('signed_transaction', signed_transaction)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        mosaic: Mosaic,
        duration: int,
        signed_transaction: SignedTransaction,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new lock funds transaction.

        :param deadline: Deadline to include transaction.
        :param mosaic: Locked mosaic.
        :param duration: Duration to lock funds.
        :param signed_transaction: Signed transaction for which funds are locked.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.LOCK,
            deadline,
            mosaic,
            duration,
            signed_transaction,
            max_fee,
        )

    @property
    def hash(self) -> str:
        """Get transaction hash from signed transaction."""
        return self.signed_transaction.hash

    @property
    def mosaic_id(self) -> typing.Union[MosaicId, NamespaceId]:
        """Get mosaic ID from mosaic."""
        return self.mosaic.id

    @property
    def amount(self) -> int:
        """Get mosaic amount from mosaic."""
        return self.mosaic.amount

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        mosaic_size = Mosaic.CATBUFFER_SIZE
        duration_size = util.U64_BYTES
        hash_size = util.U8_BYTES * 32
        return mosaic_size + duration_size + hash_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export lock funds-specific data to catbuffer."""

        # Mosaic mosaic
        # uint64 duration
        # uint8[32] hash
        mosaic = self.mosaic.to_catbuffer(network_type)
        duration = util.u64_to_catbuffer(self.duration)
        hash = util.unhexlify(self.hash)
        return mosaic + duration + hash

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load lock funds-specific data data from catbuffer."""

        # Mosaic mosaic
        # uint64 duration
        # uint8[32] hash
        mosaic, data = Mosaic.create_from_catbuffer_pair(data, network_type)
        duration = util.u64_from_catbuffer(data[:util.U64_BYTES])
        data = data[util.U64_BYTES:]
        hash = data[:util.U8_BYTES * 32]
        data = data[util.U8_BYTES * 32:]
        signed_transaction = SignedTransaction.create_from_announced(
            hash,
            TransactionType.AGGREGATE_BONDED,
            network_type
        )

        self._set('mosaic', mosaic)
        self._set('duration', duration)
        self._set('signed_transaction', signed_transaction)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'mosaicId', 'amount', 'duration', 'hash'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'mosaicId': util.u64_to_dto(int(self.mosaic_id)),
            'amount': util.u64_to_dto(self.amount),
            'duration': util.u64_to_dto(self.duration),
            'hash': self.hash,
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        mosaic_id = MosaicId(util.u64_from_dto(data['mosaicId']))
        amount = util.u64_from_dto(data['amount'])
        duration = util.u64_from_dto(data['duration'])
        signed_transaction = SignedTransaction.create_from_announced(
            data['hash'],
            TransactionType.AGGREGATE_BONDED,
            network_type
        )
        self._set('mosaic', Mosaic(mosaic_id, amount))
        self._set('duration', duration)
        self._set('signed_transaction', signed_transaction)


@register_transaction('LOCK')
class LockFundsInnerTransaction(InnerTransaction, LockFundsTransaction):
    """Embedded lock funds transaction."""

    __slots__ = ()



"""
    transaction_status_error
    ========================

    Model representing errors in transactions from listeners.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from ..account.address import Address
from .deadline import Deadline
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['TransactionStatusError']


@util.inherit_doc
@util.dataclass(frozen=True)
class TransactionStatusError(util.DTO):
    """
    Model representing errors in transactions from listeners.

    :param hash: Transaction hash.
    :param status: Status error message.
    :param deadline: Transaction deadline.

    DTO Format:
        .. code-block:: yaml

            # The DTO format is implied from the typescript SDK,
            # and is only returned by listeners.
            TransactionStatusDTO:
                status: string
                # Hex(Hash) (64-bytes)
                hash: string
                deadline: UInt64DTO
    """

    hash: str
    status: str
    deadline: Deadline
    channel_name: str
    address: Address

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys_l1 = {'hash', 'status', 'deadline', 'meta'}
        required_keys_l2 = {'channelName', 'address'}
        return (
            cls.validate_dto_required(data, required_keys_l1)
            and cls.validate_dto_all(data, required_keys_l1)
            and cls.validate_dto_required(data['meta'], required_keys_l2)
            and cls.validate_dto_all(data['meta'], required_keys_l2)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        meta = {
            'channelName': self.channel_name,
            'address': self.address.address,
        }

        return {
            'hash': self.hash,
            'status': self.status,
            'deadline': util.u64_to_dto(self.deadline.to_timestamp()),
            'meta': meta,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        meta = data['meta']
        return cls(
            hash=data['hash'],
            status=data['status'],
            deadline=Deadline.create_from_timestamp(util.u64_from_dto(data['deadline'])),
            channel_name=meta['channelName'],
            address=Address.create_from_encoded(meta['address']),
        )



"""
    account_link_transaction
    ========================

    Transaction to delegate account importance to a proxy account.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .link_action import LinkAction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = [
    'AccountLinkTransaction',
    'AccountLinkInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('LINK_ACCOUNT')
class AccountLinkTransaction(Transaction):
    """
    Account link transaction.

    Delegates account importance to a proxy account for harvesting.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param remote_account_key: Public key of the remote account.
    :param link_action: Account link action.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    remote_account_key: str
    link_action: LinkAction

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        remote_account_key: str,
        link_action: LinkAction,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        remote_account_key = util.encode_hex(remote_account_key)
        if len(remote_account_key) != 64:
            raise ValueError("Invalid remote account public key length.")
        super().__init__(
            TransactionType.LINK_ACCOUNT,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('remote_account_key', remote_account_key)
        self._set('link_action', link_action)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        remote_account_key: str,
        link_action: LinkAction,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new account link transaction.

        :param deadline: Deadline to include transaction.
        :param remote_account_key: Public key of the remote account.
        :param link_action: Account link action.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.LINK_ACCOUNT,
            deadline,
            remote_account_key,
            link_action,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        key_size = util.U8_BYTES * 32
        action_size = util.U8_BYTES
        return key_size + action_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export account link-specific data to catbuffer."""

        # uint8_t[32] remote_account_key
        # uint8_t link_action
        remote_account_key = util.unhexlify(self.remote_account_key)
        link_action = self.link_action.to_catbuffer(network_type)
        return remote_account_key + link_action

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load account link-specific data data from catbuffer."""

        # uint8_t[32] remote_account_key
        # uint8_t link_action
        remote_account_key = util.hexlify(data[:32])
        link_action = LinkAction.create_from_catbuffer(data[32:33], network_type)
        data = data[33:]

        self._set('remote_account_key', remote_account_key)
        self._set('link_action', link_action)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'remoteAccountKey', 'action'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'remoteAccountKey': self.remote_account_key,
            'action': self.link_action.to_dto(network_type),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        remote_account_key = data['remoteAccountKey']
        link_action = LinkAction.create_from_dto(data['action'])
        self._set('remote_account_key', remote_account_key)
        self._set('link_action', link_action)


@register_transaction('LINK_ACCOUNT')
class AccountLinkInnerTransaction(InnerTransaction, AccountLinkTransaction):
    """Embedded account link transaction."""

    __slots__ = ()



import binascii

from Crypto.Hash import SHA224, SHA256, SHA384, SHA512
from xpxchain.util import hashlib
from tests import harness
from .helper import *

HASH = {
    224: SHA224,
    256: SHA256,
    384: SHA384,
    512: SHA512,
}


def generate_testcase(bits, full_hexdigest):
    name = "sha{}".format(bits)
    digest_size = bits // 8
    func = getattr(hashlib, name)
    func_c = HASH[bits].new

    class Sha2Test(harness.TestCase):

        def test_init(self):
            actual = hexdigest(func(), b'Hello World!')
            expected = hexdigest(func_c(), b'Hello World!')
            self.assertEqual(actual, expected)

        def test_properties(self):
            hasher = func()
            self.assertEqual(hasher.digest_size, digest_size)

        def test_update(self):
            hasher = func()
            hasher.update(b'Hello')
            hasher.update(b' ')
            hasher.update(b'World!')
            self.assertEqual(hasher.hexdigest(), full_hexdigest)

        def test_digest(self):
            self.assertEqual(digest(func(), b'Hello World!'), binascii.unhexlify(full_hexdigest))

        def test_hexdigest(self):
            self.assertEqual(hexdigest(func(), b'Hello World!'), full_hexdigest)

        def test_bytes(self):
            for test in STRING_TESTS:
                actual = hexdigest(func(), test)
                expected = hexdigest(func_c(), test)
                self.assertEqual(actual, expected)

    return Sha2Test


TestSha3224 = generate_testcase(224, "4575bb4ec129df6380cedde6d71217fe0536f8ffc4e18bca530a7a1b")
TestSha3256 = generate_testcase(256, "7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069")
TestSha3384 = generate_testcase(384, "bfd76c0ebbd006fee583410547c1887b0292be76d582d96c242d2a792723e3fd6fd061f9d5cfd13b8f961358e6adba4a")
TestSha3512 = generate_testcase(512, "861844d6704e8573fec34d967e20bcfef3d424cf48be04e6dc08f2bd58c729743371015ead891cc3cf1c9d34b49264b510751b1ff9e537937bc46b5d6ff4ecc8")



"""
    aitertools
    ==========

    Itertools for asynchronous operations.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys


async def aenumerate(aiterable):
    """`enumerate` for asynchronous generators."""

    n = 0
    async for value in aiterable:
        yield (n, value)
        n += 1


async def aslice(aiterable, *args):
    """`islice` for asynchronous generators."""

    # Normalize the input arguments.
    slc = slice(*args)
    start = slc.start or 0
    stop = slc.stop or sys.maxsize
    assert slc.step is None or slc.step == 1

    async for (index, value) in aenumerate(aiterable):
        if index < start:
            continue
        elif index >= stop:
            break
        yield value



"""
    sync_announce
    =============

    Signed transaction to announce and sync to network.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .signed_transaction import SignedTransaction
from ..account.address import Address
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['SyncAnnounce']


@util.dataclass(frozen=True)
class SyncAnnounce(util.DTO):
    """
    Signed transaction to announce and sync.

    :param payload: Signed transaction data.
    :param hash: Transaction hash.
    :param address: Transaction address.
    """

    payload: str
    hash: str
    address: str

    def __init__(
        self,
        payload: typing.AnyStr,
        hash: typing.AnyStr,
        address: str,
    ) -> None:
        payload = util.encode_hex(payload)
        hash = util.encode_hex(hash)
        if len(hash) != 64:
            raise ValueError('Transaction hash must be 64 characters long.')
        self._set('payload', payload)
        self._set('hash', hash)
        self._set('address', address)

    @classmethod
    def create(cls, transaction: SignedTransaction):
        """
        Create sync announce object from signed transaction data.

        :param transaction: Signed transaction data.
        """
        public_key = transaction.signer
        network_type = transaction.network_type
        address = Address.create_from_public_key(public_key, network_type)

        return cls(
            payload=transaction.payload,  # type: ignore
            hash=transaction.hash,  # type: ignore
            address=address.address,
        )

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'payload', 'hash', 'address'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'payload': self.payload,
            'hash': self.hash,
            'address': self.address,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            payload=data['payload'],
            hash=data['hash'],
            address=data['address'],
        )



"""
    coroutine
    =========

    Wrapper for asynchronous coroutine.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing
import rx


class Coroutine:
    """Coroutine wrapper type."""

    def __init__(self, value, loop=None):
        self._value = value
        self._loop = loop
        self._observable = None

    def close(self) -> None:
        """Raise GeneratorExit inside coroutine."""
        self._value.close()

    def send(self, arg):
        """Send `arg` into coroutine."""
        return self._value.send(arg)

    def throw(self, exception_type, *args):
        """Raise exception in coroutine."""
        return self._value.throw(exception_type, *args)

    def to_observable(self):
        """Export awaitable object to Observable."""
        if self._observable is None:
            self._observable = rx.Observable.from_coroutine(
                self._value,
                loop=self._loop
            )
        return self._observable

    @property
    def cr_await(self):
        return self._value.cr_await

    @property
    def cr_code(self):
        return self._value.cr_code

    @property
    def cr_frame(self):
        return self._value.cr_frame

    @property
    def cr_running(self):
        return self._value.cr_running

    def __await__(self):
        if self._observable is None:
            return self._value.__await__()
        return self._observable.__await__()


class CoroutineMixin(typing.Awaitable):
    """Mixin for awaitable classes to support `to_observable`."""

    def to_observable(self):
        """Export awaitable object to Observable."""
        loop = getattr(self, 'loop', None)
        return rx.Observable.from_coroutine(
            self.__aiter__(),
            loop=loop
        )



"""
    block_info
    ==========

    Block data and metadata pair.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .block_type import BlockType
from .network_type import NetworkType, OptionalNetworkType
from ..account.public_account import PublicAccount
from ..transaction.transaction_version import TransactionVersion
from ... import util

__all__ = ['BlockInfo']

MerkleTreeType = typing.Sequence[str]
OptionalMerkleTreeType = typing.Optional[MerkleTreeType]


@util.inherit_doc
@util.dataclass(frozen=True)
class BlockInfo(util.DTO):
    """
    Basic information describing a block.

    :param hash: Block hash.
    :param generation_hash: Generation hash.
    :param total_fee: Sum of all transaction fees included in block.
    :param num_transactions: The number of transactions included in block.
    :param num_statements: The number of statements included in block.
    :param signature: Block signature.
    :param signer: Public account of the block harvester.
    :param network_type: Network type.
    :param version: Transaction version.
    :param type: Block type.
    :param height: Height at which block was confirmed.
    :param timestamp: Seconds elapsed since nemesis block.
    :param difficulty: POI difficult to harvest block.
    :param previous_block_hash: Last block hash.
    :param block_transactions_hash: Block transaction hash.
    :param block_receipts_hash: Block receipt hash.
    :param state_hash: State hash.
    :param beneficiary: (Optional) Public account of beneficiary.

    DTO Format:
        .. code-block:: yaml

            HeightDTO:
                height: UInt64DTO

            BlockMetaDTO:
                # Hex(Hash) (64-bytes)
                hash: string
                # Hex(Hash) (64-bytes)
                generationHash: string
                totalFee: UInt64DTO
                numTransactions: integer
                # Hex(Hash)[] (64-bytes)
                subCacheMerkleRoots?: string[]

            BlockDTO:
                # Hex(Signature) (128-bytes)
                signature: string
                # Hex(PublicKey) (64-bytes)
                signer: string
                version: integer
                type: integer
                height: UInt64DTO
                timestamp: UInt64DTO
                difficulty: UInt64DTO
                feeMultiplier: integer
                # Hex(Hash) (64-bytes)
                previousBlockHash: string
                # Hex(Hash) (64-bytes)
                blockTransactionsHash: string
                # Hex(Hash) (64-bytes)
                blockReceiptsHash: string
                # Hex(Hash) (64-bytes)
                stateHash: string
                # Hex(PublicKey) (64-bytes)
                beneficiaryPublicKey: string

            BlockInfoDTO:
                meta: BlockMetaDTO
                block: BlockDTO
    """

    hash: str
    generation_hash: str
    total_fee: int
    num_transactions: int
    num_statements: int
    signature: str
    signer: PublicAccount
    network_type: NetworkType
    version: TransactionVersion
    type: BlockType
    height: int
    timestamp: int
    difficulty: int
    fee_multiplier: int
    previous_block_hash: str
    block_transactions_hash: str
    block_receipts_hash: str
    state_hash: str
    beneficiary: typing.Optional[PublicAccount]
    fee_interest: int
    fee_interest_denominator: int
    merkle_tree: OptionalMerkleTreeType

    def __init__(
        self,
        hash: typing.AnyStr,
        generation_hash: typing.AnyStr,
        total_fee: int,
        num_transactions: int,
        num_statements: int,
        signature: typing.AnyStr,
        signer: PublicAccount,
        network_type: NetworkType,
        version: TransactionVersion,
        type: BlockType,
        height: int,
        timestamp: int,
        difficulty: int,
        fee_multiplier: int,
        previous_block_hash: typing.AnyStr,
        block_transactions_hash: typing.AnyStr,
        block_receipts_hash: typing.AnyStr,
        state_hash: typing.AnyStr,
        fee_interest: int,
        fee_interest_denominator: int,
        beneficiary: typing.Optional[PublicAccount] = None,
        merkle_tree: OptionalMerkleTreeType = None,
    ) -> None:
        hash = util.encode_hex(hash)
        generation_hash = util.encode_hex(generation_hash)
        signature = util.encode_hex(signature)
        previous_block_hash = util.encode_hex(previous_block_hash)
        block_transactions_hash = util.encode_hex(block_transactions_hash)
        block_receipts_hash = util.encode_base64(block_receipts_hash)
        state_hash = util.encode_hex(state_hash)
        self._set('hash', hash)
        self._set('generation_hash', generation_hash)
        self._set('total_fee', total_fee)
        self._set('num_transactions', num_transactions)
        self._set('num_statements', num_statements)
        self._set('signature', signature)
        self._set('signer', signer)
        self._set('network_type', network_type)
        self._set('version', version)
        self._set('type', type)
        self._set('height', height)
        self._set('timestamp', timestamp)
        self._set('difficulty', difficulty)
        self._set('fee_multiplier', fee_multiplier)
        self._set('previous_block_hash', previous_block_hash)
        self._set('block_transactions_hash', block_transactions_hash)
        self._set('block_receipts_hash', block_receipts_hash)
        self._set('state_hash', state_hash)
        self._set('fee_interest', fee_interest)
        self._set('fee_interest_denominator', fee_interest_denominator)
        self._set('beneficiary', beneficiary)
        self._set('merkle_tree', merkle_tree or [])

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'meta', 'block'}
        required_l21 = {
            'hash',
            'generationHash'
            # 'subCacheMerkleRoots',
            # 'totalFee',
            # 'numTransactions',
            # 'numStatements'
        }
        required_l22 = {
            'signature',
            'signer',
            'version',
            'type',
            'height',
            'timestamp',
            'difficulty',
            'previousBlockHash',
            'blockTransactionsHash'
            # 'feeMultiplier',
            # 'blockReceiptsHash',
            # 'stateHash',
            # 'beneficiary',
            # 'feeInterest',
            # 'feeInterestDenominator'
        }

        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            # Level 2_1
            and cls.validate_dto_required(data['meta'], required_l21)
            # and cls.validate_dto_all(data['meta'], required_l21)
            # Level 2_2
            and cls.validate_dto_required(data['block'], required_l22)
            # and cls.validate_dto_all(data['block'], required_l22)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        meta = {
            'hash': self.hash,
            'generationHash': self.generation_hash,
            'subCacheMerkleRoots': self.merkle_tree,
            'totalFee': util.u64_to_dto(self.total_fee),
            'numTransactions': self.num_transactions,
            'numStatements': self.num_statements,
        }
        block = {
            'signature': self.signature,
            'signer': self.signer.public_key,
            'version': int(self.version) | (int(network_type) << 24),
            'type': self.type.to_dto(network_type),
            'height': util.u64_to_dto(self.height),
            'timestamp': util.u64_to_dto(self.timestamp),
            'difficulty': util.u64_to_dto(self.difficulty),
            'feeMultiplier': util.u32_to_dto(self.fee_multiplier),
            'previousBlockHash': self.previous_block_hash,
            'blockTransactionsHash': self.block_transactions_hash,
            'blockReceiptsHash': self.block_receipts_hash,
            'stateHash': self.state_hash,
            'feeInterest': self.fee_interest,
            'feeInterestDenominator': self.fee_interest_denominator
        }

        if self.beneficiary is not None:
            public_key = util.unhexlify(self.beneficiary.public_key)
            meta['beneficiary'] = util.b64encode(public_key)

        return {
            'meta': meta,
            'block': block,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        meta = data['meta']
        block = data['block']
        version = block['version']
        network_type = NetworkType((version >> 24) & 0x000000FF)
        beneficiary = None
        if 'beneficiary' in block:
            # TODO(ahuszagh) Is base64-encoded rather than hex-encoded.
            #   Should be fixed in an upcoming version.
            beneficiary = PublicAccount.create_from_public_key(block['beneficiary'], network_type)
        return cls(
            hash=meta['hash'],
            generation_hash=meta['generationHash'],
            merkle_tree=meta.get('subCacheMerkleRoots', []),
            total_fee=util.u64_from_dto(meta.get('totalFee', [0, 0])),
            num_transactions=meta.get('numTransactions', 0),
            num_statements=meta.get('numStatements', 1),
            signature=block['signature'],
            signer=PublicAccount.create_from_public_key(block['signer'], network_type),
            version=TransactionVersion(version & 0xFFFFFF),  # Version are 3B
            type=BlockType.create_from_dto(block['type'], network_type),
            height=util.u64_from_dto(block['height']),
            timestamp=util.u64_from_dto(block['timestamp']),
            difficulty=util.u64_from_dto(block['difficulty']),
            previous_block_hash=block['previousBlockHash'],
            block_transactions_hash=block['blockTransactionsHash'],
            network_type=network_type,
            fee_multiplier=util.u32_from_dto(block.get('feeMultiplier', 0)),
            block_receipts_hash=block.get('blockReceiptsHash', ''),
            state_hash=block.get('stateHash', ''),
            beneficiary=beneficiary,
            fee_interest=block.get('feeInterest', 0),
            fee_interest_denominator=block.get('feeInterestDenominator', 0)
        )



"""
    transfer_transaction
    ====================

    Transfer transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .message import Message
from .plain_message import EMPTY_MESSAGE, PlainMessage
from .recipient import Recipient, RecipientType
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..mosaic.mosaic import Mosaic, MosaicList
from ... import util

__all__ = [
    'TransferTransaction',
    'TransferInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('TRANSFER')
class TransferTransaction(Transaction):
    """
    Transfer transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param recipient: Address or namespace ID alias of recipient.
    :param mosaics: Sequence of mosaics.
    :param message: Transaction message (up to 2048 characters).
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    recipient: RecipientType
    mosaics: MosaicList
    message: Message

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        recipient: RecipientType,
        max_fee: int = 0,
        mosaics: typing.Optional[MosaicList] = None,
        message: Message = EMPTY_MESSAGE,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None
    ) -> None:
        super().__init__(
            TransactionType.TRANSFER,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info
        )
        self._set('recipient', recipient)
        self._set('mosaics', mosaics or [])
        self._set('message', message)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        recipient: RecipientType,
        network_type: NetworkType,
        mosaics: typing.Optional[MosaicList] = None,
        message: Message = EMPTY_MESSAGE,
        max_fee: int = 0,
    ):
        """
        Create new transfer transaction.

        :param deadline: Deadline to include transaction.
        :param recipient: Address or namespace ID alias of recipient.
        :param mosaics: Sequence of mosaics.
        :param message: Transaction message (up to 2048 characters).
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.TRANSFER,
            deadline,
            recipient,
            max_fee,
            mosaics,
            message
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        # Extra 3 + address_size + message_size + $MOSAIC_SIZE * mosaics_count
        # 2 for message_size, and 1 for mosaics_count.
        extra_size = util.U8_BYTES + util.U16_BYTES
        recipient_size = Recipient.CATBUFFER_SIZE
        message_size = self.message.catbuffer_size_specific()
        mosaics_size = Mosaic.CATBUFFER_SIZE * len(self.mosaics)
        return extra_size + recipient_size + message_size + mosaics_size

    def to_mosaics_bytes(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Get the serialized byte array of all mosaics."""

        return util.Model.sequence_to_catbuffer(
            self.mosaics,
            network_type
        )

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export transfer-specific data to catbuffer."""

        # uint8_t[25] recipient
        # uint16_t message_size
        # uint8_t mosaics_count
        # uint8_t message type
        # uint8_t[message_size] message
        # Mosaic[mosaics_count] mosaics
        recipient = Recipient.to_catbuffer(self.recipient, network_type)
        message_size = util.u16_to_catbuffer(self.message.catbuffer_size_specific())
        mosaics_count = util.u8_to_catbuffer(len(self.mosaics))
        message = self.message.to_catbuffer(network_type)
        mosaics = self.to_mosaics_bytes(network_type)

        return recipient + message_size + mosaics_count + message + mosaics

    def load_mosaics_bytes(
        self,
        data: bytes,
        count: int,
        network_type: NetworkType,
    ) -> bytes:
        """Load mosaics data from catbuffer."""

        mosaics, data = Mosaic.sequence_from_catbuffer_pair(data, count, network_type)
        self._set('mosaics', mosaics)
        return data

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load transfer-specific data from catbuffer."""

        # uint8_t[25] recipient
        # uint16_t message_size
        # uint8_t mosaics_count
        # uint8_t[message_size] message
        # Mosaic[mosaics_count] mosaics
        recipient, data = Recipient.create_from_catbuffer_pair(data, network_type)
        message_size = util.u16_from_catbuffer(data[:util.U16_BYTES])
        data = data[util.U16_BYTES:]
        mosaics_count = util.u8_from_catbuffer(data[:util.U8_BYTES])
        data = data[util.U8_BYTES:]
        message = PlainMessage(data[1:message_size])
        data = data[message_size:]
        data = self.load_mosaics_bytes(data, mosaics_count, network_type)

        self._set('recipient', recipient)
        self._set('message', message)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'recipient'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'recipient': Recipient.to_dto(self.recipient, network_type),
            'mosaics': Mosaic.sequence_to_dto(self.mosaics, network_type),
            'message': self.message.to_dto(network_type),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        recipient = Recipient.create_from_dto(data['recipient'], network_type)
        mosaics = Mosaic.sequence_from_dto(data.get('mosaics', []), network_type)
        message = EMPTY_MESSAGE
        if 'message' in data:
            message = PlainMessage.create_from_dto(data['message'], network_type)

        self._set('recipient', recipient)
        self._set('mosaics', mosaics)
        self._set('message', message)


@register_transaction('TRANSFER')
class TransferInnerTransaction(InnerTransaction, TransferTransaction):
    """Embedded transfer transaction."""

    __slots__ = ()



"""
    base64
    ======

    Utilities for converting to and from base32 and base64 encodings.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import base64
import typing

__all__ = [
    'b32encode',
    'b32decode',
    'b64encode',
    'b64decode',
    'decode_base32',
    'decode_base64',
    'encode_base32',
    'encode_base64',
]


def b32encode(data: bytes, with_suffix=True) -> str:
    """Encode bytes data to a base32-encoded string."""

    encoded = base64.b32encode(data).decode('ascii')
    return remove_suffix(encoded, with_suffix=with_suffix)


def encode_base32(data: typing.AnyStr, with_suffix=True) -> str:
    """Encodes raw bytes to base32."""

    if isinstance(data, str):
        return remove_suffix(data, with_suffix=with_suffix)
    return b32encode(data, with_suffix=with_suffix)


def b32decode(data: typing.AnyStr, with_suffix=True) -> bytes:
    """Decode bytes data from a base32-encoded string."""

    if not with_suffix:
        data = add_suffix(data, 8)
    return base64.b32decode(data)


def decode_base32(data: typing.AnyStr, with_suffix=True) -> bytes:
    """Decode base32 data to raw bytes."""

    if isinstance(data, (bytes, bytearray)):
        return data
    return b32decode(data, with_suffix=with_suffix)


def b64encode(data: bytes, altchars=None, with_suffix=True) -> str:
    """Encode bytes data to a base64-encoded string."""

    encoded = base64.b64encode(data, altchars=altchars).decode('ascii')
    return remove_suffix(encoded, with_suffix=with_suffix)


def encode_base64(data: typing.AnyStr, altchars=None, with_suffix=True) -> str:
    """Encodes raw bytes to base64."""

    if isinstance(data, str):
        return remove_suffix(data, with_suffix=with_suffix)
    return b64encode(data, altchars=altchars, with_suffix=with_suffix)


def b64decode(data: typing.AnyStr, altchars=None, with_suffix=True) -> bytes:
    """Decode bytes data from a base64-encoded string."""

    if not with_suffix:
        data = add_suffix(data, 4)
    return base64.b64decode(data, altchars=altchars)


def decode_base64(data: typing.AnyStr, altchars=None, with_suffix=True) -> bytes:
    """Decode base64 data to raw bytes."""

    if isinstance(data, (bytes, bytearray)):
        return data
    return b64decode(data, altchars=altchars, with_suffix=with_suffix)


@typing.no_type_check
def add_suffix(data: typing.AnyStr, width: int) -> typing.AnyStr:
    """Add basen suffix to string ('=')."""

    if isinstance(data, str):
        padchar = '='
    else:
        padchar = b'='

    trailing = len(data) % width
    if trailing != 0:
        pad = width - trailing
        data = data + padchar * pad

    return data


def remove_suffix(data: str, with_suffix=True) -> str:
    """Remove '=' suffix from encoded str."""

    if not with_suffix:
        return data.rstrip('=')
    return data



"""
    public_account
    ==============

    Account public key and address.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .address import Address
from ..blockchain.network_type import NetworkType
from ... import util
from ...util.signature import ed25519

__all__ = ['PublicAccount']


@util.inherit_doc
@util.dataclass(frozen=True)
class PublicAccount(util.Object):
    """
    Describe public account information via public key and account address.

    :param address: Address for the account.
    :param public_key: Hex-encoded public key (with or without '0x' prefix).
    """

    address: Address
    public_key: str
    CATBUFFER_SIZE: typing.ClassVar[int] = 32 * util.U8_BYTES

    def __init__(
        self,
        address: Address,
        public_key: str,
    ) -> None:
        public_key = util.encode_hex(public_key)
        if len(public_key) != 64:
            raise ValueError("Invalid public key length")
        self._set('address', address)
        self._set('public_key', public_key)

    @property
    def network_type(self) -> NetworkType:
        """Get network type."""
        return self.address.network_type

    @classmethod
    def create_from_public_key(
        cls,
        public_key: typing.AnyStr,
        network_type: NetworkType,
    ):
        """
        Create PublicAccount from the public key and network type.

        :param public_key: Hex-encoded or raw bytes for public key.
        :param network_type: Network type for address.
        :return: PublicAccount object.
        """

        public_key = util.encode_hex(public_key)
        address = Address.create_from_public_key(public_key, network_type)

        return cls(address, public_key)

    def verify_signature(
        self,
        data: typing.AnyStr,
        signature: typing.AnyStr
    ) -> bool:
        """
        Verify a signature.

        :param data: Hex-encoded or raw bytes used to generate signature.
        :param public_key: Hex-encoded or raw bytes for signature.
        :return: Boolean representing if the signature was verified.
        """

        data = util.decode_hex(data, with_prefix=True)
        signature = util.decode_hex(signature, with_prefix=True)
        if len(signature) != 64:
            raise ValueError("Signature length is incorrect.")

        public_key = util.unhexlify(self.public_key)
        key = ed25519.sha3.VerifyingKey(public_key)
        try:
            key.verify(signature, data)
            return True
        except ed25519.sha3.BadSignatureError:
            return False

    def verify_transaction(
        self,
        transaction: typing.AnyStr
    ) -> bool:
        """
        Verify signed transaction data.

        :param transaction: Hex-encoded or raw bytes for transaction data.
        :return: Boolean representing if the transaction signature was verified.
        """

        transaction = util.decode_hex(transaction, with_prefix=True)

        # Skip first 100 bytes.
        # uint32_t size
        # uint8_t[64] signature
        # uint8_t[32] signer
        data = transaction[100:]
        signature = transaction[4:68]
        return self.verify_signature(data, signature)


PublicAccountList = typing.Sequence[PublicAccount]



"""
    cosignature_transaction
    =======================

    Aggregate transaction to cosign.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .aggregate_transaction import AggregateTransaction
from .cosignature_signed_transaction import CosignatureSignedTransaction
from .transaction_info import TransactionInfo
from ..account.account import Account
from ... import util

__all__ = [
    'CosignatureTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
class CosignatureTransaction(util.Object):
    """Cosignature transaction is used to sign aggregate transactions."""

    transaction: AggregateTransaction

    def __init__(
        self,
        transaction: AggregateTransaction
    ) -> None:
        if transaction.is_unannounced():
            raise ValueError('Transaction to cosign must be announced first.')
        self._set('transaction', transaction)

    @classmethod
    def create(
        cls,
        transaction: AggregateTransaction
    ):
        """
        Create cosignature transaction.

        :param account: Account to sign with.
        """
        return cls(transaction)

    def sign_with(self, account: Account) -> CosignatureSignedTransaction:
        """
        Serialize and sign transaction.

        :param account: Account to sign with.
        """

        transaction_info = self.transaction.transaction_info
        if transaction_info is None:
            raise ValueError('Transaction info not present.')
        parent_hash = typing.cast(TransactionInfo, transaction_info).hash
        if parent_hash is None:
            raise ValueError('Transaction info to cosign has no hash.')

        signature = util.hexlify(account.sign_data(parent_hash))
        signer = account.public_key
        return CosignatureSignedTransaction(parent_hash, signature, signer)



import sys

from xpxchain import util
from tests import harness

HAS_REACTIVE = 'rx' in sys.modules


@util.observable
async def foo():
    return 3


@util.observable
class Bar:
    def __await__(self):
        return foo().__await__()


@util.observable
async def gen():
    for i in range(5):
        yield i


@util.observable
class AsyncIter:
    def __init__(self):
        self.i = 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.i == 5:
            raise StopAsyncIteration
        value = self.i
        self.i += 1
        return value


class TestObservableDecorator(harness.TestCase):

    async def test_coroutine(self):
        self.assertEqual(await foo(), 3)
        self.assertEqual(await Bar(), 3)

    async def test_coroutine_subscribe(self):
        src = foo()
        if HAS_REACTIVE:
            src = src.to_observable()
            src.subscribe(lambda x: None)
        self.assertEqual(await src, 3)

    async def test_async_generator(self):
        l = [i async for i in gen()]
        self.assertEqual(l, list(range(5)))
        if HAS_REACTIVE:
            l = [i async for i in gen().to_observable()]
            self.assertEqual(l, list(range(5)))

    async def test_async_iterator(self):
        l = [i async for i in AsyncIter()]
        self.assertEqual(l, list(range(5)))

    async def test_subscribe(self):
        l1 = []
        l2 = []
        src1 = gen()
        src2 = AsyncIter()

        if HAS_REACTIVE:
            obs1 = src1.to_observable()
            obs2 = src2.to_observable()
            obs1.subscribe(lambda x: l1.append(x))
            obs2.subscribe(lambda x: l2.append(x))
            await obs1
            await obs2
        else:
            l1 = [i async for i in src1]
            l2 = [i async for i in src2]

        self.assertEqual(l1, list(range(5)))
        self.assertEqual(l2, list(range(5)))



"""
    address_alias_transaction
    =========================

    Address alias transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.address import Address
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..namespace.alias_action_type import AliasActionType
from ..namespace.namespace_id import NamespaceId
from ... import util

__all__ = [
    'AddressAliasTransaction',
    'AddressAliasInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('ADDRESS_ALIAS')
class AddressAliasTransaction(Transaction):
    """
    Address alias transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param action_type: Alias action type.
    :param namespace_id: Resulting namespace ID that will be an alias.
    :param address: Address to be aliased.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    action_type: AliasActionType
    namespace_id: NamespaceId
    address: Address

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        action_type: AliasActionType,
        namespace_id: NamespaceId,
        address: Address,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        super().__init__(
            TransactionType.ADDRESS_ALIAS,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('action_type', action_type)
        self._set('namespace_id', namespace_id)
        self._set('address', address)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        action_type: AliasActionType,
        namespace_id: NamespaceId,
        address: Address,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new address alias transaction.

        :param deadline: Deadline to include transaction.
        :param action_type: Alias action type.
        :param namespace_id: Resulting namespace ID that will be an alias.
        :param address: Address to be aliased.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.ADDRESS_ALIAS,
            deadline,
            action_type,
            namespace_id,
            address,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        action_type_size = AliasActionType.CATBUFFER_SIZE
        namespace_id_size = util.U64_BYTES
        address_size = 25 * util.U8_BYTES
        return action_type_size + namespace_id_size + address_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export address alias-specific data to catbuffer."""

        # uint8_t action_type
        # uint64_t namespace_id
        # Address address
        action_type = self.action_type.to_catbuffer(network_type)
        namespace_id = util.u64_to_catbuffer(int(self.namespace_id))
        address = self.address.encoded

        return action_type + namespace_id + address

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load address alias-specific data data from catbuffer."""

        # uint8_t action_type
        # uint64_t namespace_id
        # Address address
        action_type = AliasActionType.create_from_catbuffer(data, network_type)
        namespace_id = NamespaceId(util.u64_from_catbuffer(data[1:9]))
        address = Address.create_from_encoded(data[9:34])
        data = data[34:]

        self._set('action_type', action_type)
        self._set('namespace_id', namespace_id)
        self._set('address', address)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'aliasAction', 'namespaceId', 'address'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'aliasAction': self.action_type.to_dto(network_type),
            'namespaceId': util.u64_to_dto(int(self.namespace_id)),
            'address': util.hexlify(self.address.encoded),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        action_type = AliasActionType.create_from_dto(data['aliasAction'], network_type)
        namespace_id = NamespaceId(util.u64_from_dto(data['namespaceId']))
        address = Address.create_from_encoded(data['address'])

        self._set('action_type', action_type)
        self._set('namespace_id', namespace_id)
        self._set('address', address)


@register_transaction('ADDRESS_ALIAS')
class AddressAliasInnerTransaction(InnerTransaction, AddressAliasTransaction):
    """Embedded address alias transaction."""

    __slots__ = ()



"""
    mosaic
    ======

    Description of an asset.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .mosaic_id import MosaicId
from ..blockchain.network_type import OptionalNetworkType
from ..namespace.namespace_id import NamespaceId
from ... import util

__all__ = ['Mosaic']

IdType = typing.Union[MosaicId, NamespaceId]


@util.inherit_doc
@util.dataclass(frozen=True, amount=0)
class Mosaic(util.Model):
    """
    Basic information describing a mosaic.

    :param id: Identifier for mosaic.
    :param amount: Mosaic quantity in the smallest unit possible.

    DTO Format:
        .. code-block:: yaml

            MosaicDTO:
                id: UInt64DTO
                amount: UInt64DTO
    """

    id: IdType
    amount: int
    CATBUFFER_SIZE: typing.ClassVar[int] = 2 * util.U64_BYTES

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'id', 'amount'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'amount': util.u64_to_dto(self.amount),
            'id': util.u64_to_dto(int(self.id)),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        mosaic_id = MosaicId(util.u64_from_dto(data['id']))
        amount = util.u64_from_dto(data['amount'])
        return cls(mosaic_id, amount)

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: typing.Optional[util.FeeCalculationStrategy] = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        mosaic_id = util.u64_to_catbuffer(int(self.id))
        amount = util.u64_to_catbuffer(self.amount)
        return mosaic_id + amount

    @classmethod
    def create_from_catbuffer(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        mosaic_id = MosaicId(util.u64_from_catbuffer(data[:8]))
        amount = util.u64_from_catbuffer(data[8:16])
        return cls(mosaic_id, amount)


MosaicList = typing.Sequence[Mosaic]



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


@harness.mocked_http_test_case({
    'clients': (client.NamespaceHTTP, client.AsyncNamespaceHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_namespace',
            'response': responses.NAMESPACE['nem'],
            'params': [models.NamespaceId.create_from_hex('84b3552d375ffa4b')],
            'method': 'get_namespace',
            'validation': [
                lambda x: (x.active, True),
                lambda x: (x.index, 0),
                lambda x: (x.meta_id, '5C7C07005CC1FE000176FA2B'),
                lambda x: (x.type, models.NamespaceType.ROOT_NAMESPACE),
                lambda x: (x.depth, 1),
                lambda x: (x.levels[0].encoded, '4bfa5f372d55b384'),
                lambda x: (x.parent_id, models.NamespaceId(0)),
                lambda x: (x.owner.address.address, 'SD3MA6SM7GWRX4DEJVAZEGFXF7G7D36MA6TMSIBM'),
                lambda x: (x.owner.public_key, '7A562888C7AE1E082579951D6D93BF931DE979360ACCA4C4085D754E5E122808'),
                lambda x: (x.start_height, 1),
                lambda x: (x.end_height, 18446744073709551615),
                lambda x: (x.alias, models.EmptyAlias()),
            ]
        },
        {
            'name': 'test_get_namespaces_name',
            'response': responses.NAMESPACE_NAMES['nem'],
            'params': [[models.NamespaceId.create_from_hex('84b3552d375ffa4b')]],
            'method': 'get_namespaces_name',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].namespace_id.encoded, '4bfa5f372d55b384'),
                lambda x: (x[0].name, 'nem'),
            ]
        },
        {
            'name': 'test_get_namespaces_from_account',
            'response': responses.NAMESPACES["nem"],
            'params': [models.Address('SD3MA6SM7GWRX4DEJVAZEGFXF7G7D36MA6TMSIBM')],
            'method': 'get_namespaces_from_account',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].meta_id, '5C7C07005CC1FE000176FA2B'),
            ]
        },
        {
            'name': 'test_get_namespaces_from_accounts',
            'response': responses.NAMESPACES["nem"],
            'params': [[models.Address('SD3MA6SM7GWRX4DEJVAZEGFXF7G7D36MA6TMSIBM')]],
            'method': 'get_namespaces_from_accounts',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].meta_id, '5C7C07005CC1FE000176FA2B'),
            ]
        },
        {
            'name': 'test_get_linked_mosaic_id',
            'response': responses.NAMESPACE["nem"],
            'params': [models.NamespaceId.create_from_hex("84b3552d375ffa4b")],
            'method': 'get_linked_mosaic_id',
            'error': ValueError,
        },
        {
            'name': 'test_get_linked_address',
            'response': responses.NAMESPACE["nem"],
            'params': [models.NamespaceId.create_from_hex("84b3552d375ffa4b")],
            'method': 'get_linked_address',
            'error': ValueError,
        },
    ],
})
class TestNamespaceHTTP(harness.TestCase):
    pass



"""
    fallback
    ========

    Fallback ed25519 implementation if a C implementation is not installed.

    Optimized version of the reference implementation of Ed25519

    Written in 2011? by Daniel J. Bernstein <djb@cr.yp.to>
               2013 by Donald Stufft <donald@stufft.io>
               2013 by Alex Gaynor <alex.gaynor@gmail.com>
               2013 by Greg Price <price@mit.edu>

    Modified in 2019 by Alex Huszagh <ahuszagh@gmail.com>

    .. warning::
        This code is not safe for use with secret keys or secret data.
        This code is susceptible to timing and side-channel attacks,
        since it uses Python's arbitrary-precision integer arithmetic,
        which may lead to the disclosure of the secrets.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import os
import typing
import warnings

from xpxchain.util import bit

# EXCEPTIONS

HashFuncType = typing.Callable[[bytes], bytes]
SECRET_LEAK_MSG = (
    'Security warning: {} using insecure ed25519'  # nosec
    ' implementation, secrets may be leaked.'
)


class SecretsWarning(UserWarning):
    pass


# API


def generate_api(hash512: HashFuncType):        # noqa: C901
    """Generate the ed25519 API from a hash function."""

    class BadSignatureError(Exception):
        pass

    class SigningKey:
        """Signing (secret) key."""

        def __init__(self, signing_key: bytes):
            if len(signing_key) == 64:
                self._signing_key = signing_key
            elif len(signing_key) == 32:
                self._signing_key = publickey(signing_key, hash512)[0]
            else:
                raise ValueError("Invalid signing key or seed length.")

        def __eq__(self, other):
            if not isinstance(other, SigningKey):
                return False
            return self._signing_key == other._signing_key

        def to_bytes(self) -> bytes:
            """Export signing key to bytes."""

            return self._signing_key

        def to_seed(self) -> bytes:
            """Export seed from signing key."""

            return self._signing_key[:32]

        def get_verifying_key(self) -> 'VerifyingKey':
            """Get verifying key from signing key."""

            return VerifyingKey(self._signing_key[32:])

        def sign(self, message: bytes) -> bytes:
            """Sign message and return signature."""

            return sign(message, self._signing_key, hash512)

        __hash__ = None

    class VerifyingKey:
        """Verifying (public) key."""

        def __init__(self, verifying_key: bytes):
            if len(verifying_key) != 32:
                raise ValueError("Invalid verifying key length.")
            self._verifying_key = verifying_key

        def __eq__(self, other):
            if not isinstance(other, VerifyingKey):
                return False
            return self._verifying_key == other._verifying_key

        def to_bytes(self) -> bytes:
            """Export verifying key to bytes."""

            return self._verifying_key

        def verify(self, signature: bytes, message: bytes) -> None:
            """Verify signature from message."""

            if not verify(self._verifying_key, signature, message, hash512):
                raise BadSignatureError("Bad signature")

        __hash__ = None

    def create_keypair(entropy=os.urandom):
        """Generate signing/verifying keypair from entropy."""

        signing_key = SigningKey(entropy(32))
        verifying_key = signing_key.get_verifying_key()
        return signing_key, verifying_key

    return BadSignatureError, SigningKey, VerifyingKey, create_keypair


# UTILITY

Point = typing.Tuple[int, int, int, int]

BITS: int = 256
Q: int = 2 ** 255 - 19
L: int = 2 ** 252 + 27742317777372353535851937790883648493


def pow2(x: int, p: int) -> int:
    """== pow(x, 2**p, q)"""

    while p > 0:
        x = x * x % Q
        p -= 1
    return x


def zero_msb(s: bytes) -> bytes:
    """Zero the most-significant bit in an array."""

    s = bytearray(s)
    s[31] = bit.clear(s[31], 7)
    return bytes(s)


def hint(m: bytes, hash512: HashFuncType) -> int:
    return int.from_bytes(hash512(m), 'little')


def inv(z: int) -> int:
    """$= z^{-1} \\mod q$, for z != 0"""

    # Adapted from curve25519_athlon.c in djb's Curve25519.
    z2 = z * z % Q                                  # 2
    z9 = pow2(z2, 2) * z % Q                        # 9
    z11 = z9 * z2 % Q                               # 11
    z2_5_0 = (z11 * z11) % Q * z9 % Q               # 31 == 2^5 - 2^0
    z2_10_0 = pow2(z2_5_0, 5) * z2_5_0 % Q          # 2^10 - 2^0
    z2_20_0 = pow2(z2_10_0, 10) * z2_10_0 % Q       # ...
    z2_40_0 = pow2(z2_20_0, 20) * z2_20_0 % Q
    z2_50_0 = pow2(z2_40_0, 10) * z2_10_0 % Q
    z2_100_0 = pow2(z2_50_0, 50) * z2_50_0 % Q
    z2_200_0 = pow2(z2_100_0, 100) * z2_100_0 % Q
    z2_250_0 = pow2(z2_200_0, 50) * z2_50_0 % Q     # 2^250 - 2^0
    return pow2(z2_250_0, 5) * z11 % Q              # 2^255 - 2^5 + 11 = q - 2


D: int = -121665 * inv(121666) % Q
I: int = pow(2, (Q - 1) // 4, Q)


def xrecover(y: int) -> int:
    xx: int = (y * y - 1) * inv(D * y * y + 1)
    x: int = pow(xx, (Q + 3) // 8, Q)

    if (x * x - xx) % Q != 0:
        x = (x * I) % Q

    if x % 2 != 0:
        x = Q - x

    return x


BY: int = 4 * inv(5)
BX: int = xrecover(BY)
B: Point = (BX % Q, BY % Q, 1, (BX * BY) % Q)
IDENT: Point = (0, 1, 1, 0)


def edwards_add(p: Point, q: Point) -> Point:
    # This is formula sequence 'addition-add-2008-hwcd-3' from
    # http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended-1.html
    (x1, y1, z1, t1) = p
    (x2, y2, z2, t2) = q

    a: int = (y1 - x1) * (y2 - x2) % Q
    b: int = (y1 + x1) * (y2 + x2) % Q
    c: int = t1 * 2 * D * t2 % Q
    d: int = z1 * 2 * z2 % Q
    e: int = b - a
    f: int = d - c
    g: int = d + c
    h: int = b + a
    x3: int = e * f
    y3: int = g * h
    t3: int = e * h
    z3: int = f * g

    return (x3 % Q, y3 % Q, z3 % Q, t3 % Q)


def edwards_double(p: Point) -> Point:
    # This is formula sequence 'dbl-2008-hwcd' from
    # http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended-1.html
    (x1, y1, z1, t1) = p

    a: int = x1 * x1 % Q
    b: int = y1 * y1 % Q
    c: int = 2 * z1 * z1 % Q
    # d: int = -a
    e: int = ((x1 + y1) * (x1 + y1) - a - b) % Q
    g: int = -a + b  # d + b
    f: int = g - c
    h: int = -a - b  # d - b
    x3: int = e * f
    y3: int = g * h
    t3: int = e * h
    z3: int = f * g

    return (x3 % Q, y3 % Q, z3 % Q, t3 % Q)


def scalarmult(p: Point, e: int) -> Point:
    if e == 0:
        return IDENT
    q: Point = scalarmult(p, e // 2)
    q = edwards_double(q)
    if e & 1:
        q = edwards_add(q, p)
    return q


def make_bpow() -> typing.List[Point]:
    """BPOW[i] == scalarmult(B, 2**i)"""

    bpow = []
    p: Point = B
    for i in range(253):
        bpow.append(p)
        p = edwards_double(p)

    return bpow


BPOW = make_bpow()


def scalarmult_b(e: int) -> Point:
    """Implements scalarmult(B, e) more efficiently."""
    # scalarmult(B, L) is the identity
    e = e % L
    p = IDENT
    for i in range(253):
        if e & 1:
            p = edwards_add(p, BPOW[i])
        e = e // 2

    assert e == 0
    return p


def encode_uint256(y: int) -> bytes:
    """Encode 256-bit integer to little-endian buffer."""

    return y.to_bytes(32, 'little')


def encode_point(p: Point) -> bytes:
    """Encode point along the elliptical curve."""

    (x, y, z, t) = p
    zi = inv(z)
    x = (x * zi) % Q
    y = (y * zi) % Q

    s = bytearray(encode_uint256(y))
    s[31] = bit.assign(s[31], 7, x & 1 == 1)

    return bytes(s)


def is_on_curve(p: Point) -> bool:
    (x, y, z, t) = p
    return (
        z % Q != 0
        and x * y % Q == z * t % Q
        and (y * y - x * x - z * z - D * t * t) % Q == 0
    )


def decode_uint256(s: bytes) -> int:
    """Decode 256-bit integer from little-endian buffer."""

    assert len(s) == 32
    return int.from_bytes(s, 'little')


def decode_point(s: bytes) -> Point:
    """Decode point along the elliptical curve."""

    assert len(s) == 32
    y = decode_uint256(zero_msb(s))
    x = xrecover(y)
    if x & 1 != bit.get(s[31], 7):
        x = Q - x
    p = (x, y, 1, (x * y) % Q)
    if not is_on_curve(p):
        raise ValueError("decoding point that is not on curve")
    return p


def decode_hash512(h: bytes) -> int:
    """Decode 512-bit hash to 256-bit integer"""

    assert len(h) == 64

    # Modify the hash buffer to clear bits prior to encoding.
    h = bytearray(h[:32])
    h[0] &= 248
    h[31] &= 127
    h[31] |= 64

    return decode_uint256(h)


def publickey(
    seed: bytes,
    hash512: HashFuncType
) -> typing.Tuple[bytes, bytes]:
    """Generate public and private key from seed."""

    warning_msg = SECRET_LEAK_MSG.format('generating verifying key')
    warnings.warn(warning_msg, SecretsWarning)

    assert len(seed) == 32

    h = hash512(seed)
    a = decode_hash512(h)
    public_key: bytes = encode_point(scalarmult_b(a))
    return seed + public_key, public_key


def sign(
    message: bytes,
    signing_key: bytes,
    hash512: HashFuncType
) -> bytes:
    """Sign message using public and private key."""

    warning_msg = SECRET_LEAK_MSG.format('signing message')
    warnings.warn(warning_msg, SecretsWarning)

    if len(signing_key) != 64:
        raise ValueError("Invalid signing key length.")

    seed = signing_key[:32]
    verifying_key = signing_key[32:64]
    h = hash512(seed)
    a = decode_hash512(h)
    r = hint(h[32:64] + message, hash512)
    p = scalarmult_b(r)
    s = (r + hint(encode_point(p) + verifying_key + message, hash512) * a) % L
    return encode_point(p) + encode_uint256(s)


def verify(
    verifying_key: bytes,
    signature: bytes,
    message: bytes,
    hash512: HashFuncType
) -> bool:
    """
    Verify signature from public key and message.

    :param verifying_key: Key to verify signature.
    :param signature: Signature generated from message.
    :param message: Original message.
    """

    warning_msg = SECRET_LEAK_MSG.format('verifying signature')
    warnings.warn(warning_msg, SecretsWarning)

    if len(verifying_key) != 32:
        raise ValueError("Invalid verifying key length.")
    if len(signature) != 64:
        raise ValueError("Invalid signature length.")

    r = decode_point(signature[:32])
    a = decode_point(verifying_key)
    s = decode_uint256(signature[32: 64])
    h = hint(encode_point(r) + verifying_key + message, hash512)

    (x1, y1, z1, t1) = p = scalarmult_b(s)
    (x2, y2, z2, t2) = q = edwards_add(r, scalarmult(a, h))

    return (
        is_on_curve(p)
        and is_on_curve(q)
        and (x1 * z2 - x2 * z1) % Q == 0
        and (y1 * z2 - y2 * z1) % Q == 0
    )



"""
    transaction_status
    ==================

    Basic information describing announced transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline, TIMESTAMP_NEMESIS_BLOCK, TIMESTAMP_NEMESIS_BLOCK_DTO
from .transaction_status_group import TransactionStatusGroup
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['TransactionStatus']


@util.inherit_doc
@util.dataclass(frozen=True)
class TransactionStatus(util.DTO):
    """
    Basic information describing announced transaction.

    :param group: Transaction status group.
    :param status: Status information describing error or success.
    :param hash: Transaction hash (hex-encoded).
    :param deadline: Transaction deadline.
    :param height: Block height at which it was confirmed or rejected.

    DTO Format:
        .. code-block:: yaml

            TransactionStatusDTO:
                group?: string
                status: string
                # Hex(Hash) (64-bytes)
                hash?: string
                deadline?: UInt64DTO
                height?: UInt64DTO
    """

    group: TransactionStatusGroup
    status: str
    hash: str
    deadline: Deadline
    height: int

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'status'}
        all_keys = required_keys | {'group', 'hash', 'deadline', 'height'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, all_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        data: dict = {'status': self.status}
        timestamp = self.deadline.to_timestamp()
        if self.group != TransactionStatusGroup.UNKNOWN:
            data['group'] = typing.cast(str, self.group.value)
        if self.hash:
            data['hash'] = self.hash
        if timestamp != TIMESTAMP_NEMESIS_BLOCK:
            data['deadline'] = util.u64_to_dto(timestamp)
        if self.height:
            data['height'] = util.u64_to_dto(self.height)

        return data

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        group = data.get('group', 'unknown')
        hash = data.get('hash', '')
        timestamp = data.get('deadline', TIMESTAMP_NEMESIS_BLOCK_DTO)
        height = data.get('height', [0, 0])
        return cls(
            group=TransactionStatusGroup(group),
            status=data['status'],
            hash=hash,
            deadline=Deadline.create_from_timestamp(util.u64_from_dto(timestamp)),
            height=util.u64_from_dto(height),
        )



# Globs
from .abc import *
from .asynchronous import *
from .base64 import *
from .binascii import *
from .dataclasses import *
from .documentation import *
from .identifier import *
from .mixin import *
from .reactive import *
from .stdint import *
from .fee import *

# Modules
#   bit
#   hashlib
#   signature

__all__ = (
    abc.__all__
    + asynchronous.__all__
    + base64.__all__
    + binascii.__all__
    + dataclasses.__all__
    + documentation.__all__
    + identifier.__all__
    + mixin.__all__
    + reactive.__all__
    + stdint.__all__
    + fee.__all__
)



import binascii
import math

from xpxchain import util
from tests import harness


class TestBase64(harness.TestCase):

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_base32_encoding(self, decoded: bytes):
        encoded = util.b32encode(decoded)
        self.assertEqual(len(encoded), 8 * int(math.ceil(len(decoded) / 5)))
        self.assertEqual(util.b32decode(encoded), decoded)

        with self.assertRaises(binascii.Error):
            util.b32decode(encoded + '0')
        with self.assertRaises(binascii.Error):
            util.b32decode(encoded[:-1])

    @harness.randomize(calls=100, data={'min_length': 0, 'max_length': 100})
    def test_base64_encoding(self, decoded: bytes):
        encoded = util.b64encode(decoded)
        self.assertEqual(len(encoded), 4 * int(math.ceil(len(decoded) / 3)))
        self.assertEqual(util.b64decode(encoded), decoded)

        with self.assertRaises(binascii.Error):
            util.b32decode(encoded + '0')
        with self.assertRaises(binascii.Error):
            util.b32decode(encoded[:-1])



from xpxchain import util
from tests import harness


class TestUMixin(harness.TestCase):

    def test_u8mixin(self):
        class Mixin(util.U8Mixin, int):
            __slots__ = ()

        value = Mixin(0x7F)
        dto = 0x7F
        catbuffer = b'\x7f'
        self.assertEqual(value.to_dto(None), dto)
        self.assertEqual(Mixin.create_from_dto(dto, None), value)
        self.assertEqual(value.to_catbuffer(None), catbuffer)
        self.assertEqual(Mixin.create_from_catbuffer(catbuffer, None), value)

    def test_u16mixin(self):
        class Mixin(util.U16Mixin, int):
            __slots__ = ()

        value = Mixin(0x7FFF)
        dto = 0x7FFF
        catbuffer = b'\xff\x7f'
        self.assertEqual(value.to_dto(None), dto)
        self.assertEqual(Mixin.create_from_dto(dto, None), value)
        self.assertEqual(value.to_catbuffer(None), catbuffer)
        self.assertEqual(Mixin.create_from_catbuffer(catbuffer, None), value)

    def test_u32mixin(self):
        class Mixin(util.U32Mixin, int):
            __slots__ = ()

        value = Mixin(0x7FFFFFFF)
        dto = 0x7FFFFFFF
        catbuffer = b'\xff\xff\xff\x7f'
        self.assertEqual(value.to_dto(None), dto)
        self.assertEqual(Mixin.create_from_dto(dto, None), value)
        self.assertEqual(value.to_catbuffer(None), catbuffer)
        self.assertEqual(Mixin.create_from_catbuffer(catbuffer, None), value)

    def test_u64mixin(self):
        class Mixin(util.U64Mixin, int):
            __slots__ = ()

        value = Mixin(0x7FFFFFFFFFFFFFFF)
        dto = [0xFFFFFFFF, 0x7FFFFFFF]
        catbuffer = b'\xff\xff\xff\xff\xff\xff\xff\x7f'
        self.assertEqual(value.to_dto(None), dto)
        self.assertEqual(Mixin.create_from_dto(dto, None), value)
        self.assertEqual(value.to_catbuffer(None), catbuffer)
        self.assertEqual(Mixin.create_from_catbuffer(catbuffer, None), value)

    def test_u128mixin(self):
        class Mixin(util.U128Mixin, int):
            __slots__ = ()

        value = Mixin(0x7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF)
        dto = [[0xFFFFFFFF, 0xFFFFFFFF], [0xFFFFFFFF, 0x7FFFFFFF]]
        catbuffer = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f'
        self.assertEqual(value.to_dto(None), dto)
        self.assertEqual(Mixin.create_from_dto(dto, None), value)
        self.assertEqual(value.to_catbuffer(None), catbuffer)
        self.assertEqual(Mixin.create_from_catbuffer(catbuffer, None), value)




from __future__ import annotations

import typing

from .metadata_type import MetadataType
from ..account.address import Address
from ..blockchain.network_type import OptionalNetworkType
from .field import Field
from ... import util

__all__ = ['AddressMetadata']


@util.inherit_doc
@util.dataclass(frozen=True)
class AddressMetadata(util.DTO):
    """

    :param metadataType:  Metadata type.
    :param fields:  Sequence of FieldDTOs.
    :param metadataId:  Metadata ID.

    DTO Format:
        .. code-block:: yaml

            MetadataDTO:
                metedataType: int
                fields: FieldDTO[]
                metedataId: string
    """

    metadata_type: MetadataType
    flds: typing.Sequence[Field]
    metadata_id: Address

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {
            'metadataType',
            'fields',
            'metadataId'
        }

        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'metadataType': util.u8_to_dto(self.metadata_type),
            'fields': [x.to_dto() for x in self.flds],
            'metadataId': self.metadata_id.plain(),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            metadata_type=MetadataType.create_from_dto(data["metadataType"]),
            flds=[Field.create_from_dto(f) for f in data["fields"]],
            metadata_id=Address.create_from_encoded(data["metadataId"])
        )



"""
    empty_alias
    ===========

    Empty alias without data.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .alias import Alias, dto_to_kwds
from .alias_type import AliasType
from ..blockchain.network_type import OptionalNetworkType

__all__ = ['EmptyAlias']


class EmptyAlias(Alias):
    """Empty alias without data."""

    __slots__ = ()

    def __init__(self, value: None = None) -> None:
        super().__init__(AliasType.NONE, value)

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        kwds = dto_to_kwds(data, network_type)
        if kwds.pop('type') != AliasType.NONE:
            raise ValueError('Alias type is not a none.')
        return cls(**kwds)



"""
    modify_mosaic_metadata_transaction
    =========================

    Modify address metadata transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..mosaic.mosaic import MosaicId
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ..metadata.metadata_type import MetadataType
from ..metadata.metadata_modification_type import MetadataModificationType
from ..metadata.metadata_modification import MetadataModification
from ... import util


MetadataModificationList = typing.Sequence[MetadataModification]

__all__ = [
    'ModifyMosaicMetadataTransaction',
    'ModifyMosaicMetadataInnerTransaction',
]

TYPES = (
    TransactionType.MODIFY_MOSAIC_METADATA,
)


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('MODIFY_MOSAIC_METADATA')
class ModifyMosaicMetadataTransaction(Transaction):
    """
    Modify mosaic metadata transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param metadata_type: type of the metadata
    :param metadata_id: mosaic ID for which metadata is modified (used only when address is not specified)
    :param modifications: the modification to make
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.

    """

    metadata_type: MetadataType
    metadata_id: MosaicId
    modifications: MetadataModificationList

    def __init__(
        self,
        network_type: NetworkType,
        type: TransactionType,
        version: TransactionVersion,
        deadline: Deadline,
        metadata_type: MetadataType,
        metadata_id: MosaicId,
        modifications: MetadataModificationList,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        if type not in TYPES:
            raise ValueError('Invalid transaction type.')
        super().__init__(
            type,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('metadata_type', metadata_type)
        self._set('metadata_id', metadata_id)
        self._set('modifications', modifications)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        metadata_type: MetadataType,
        metadata_id: MosaicId,
        modifications: MetadataModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new account metadata modification transaction.

        :param deadline: Deadline to include transaction.
        :param metadata_type: type of the metadata
        :param metadata_id: mosaic ID for which metadata is modified
        :param modifications: the modification to make
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionType.MODIFY_MOSAIC_METADATA,
            TransactionVersion.MODIFY_MOSAIC_METADATA,
            deadline,
            metadata_type,
            metadata_id,
            modifications,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        metadata_type_size = MetadataModificationType.CATBUFFER_SIZE
        metadata_id_size = util.U64_BYTES
        modifications_size = sum([x.catbuffer_size_specific() for x in self.modifications])
        return metadata_type_size + metadata_id_size + modifications_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export address alias-specific data to catbuffer."""

        # uint8_t metadata_type
        # metadata_id

        metadata_type = util.u8_to_catbuffer(self.metadata_type)
        metadata_id = util.u64_to_catbuffer(int(self.metadata_id))
        modifications = b''.join([x.to_catbuffer_specific(network_type) for x in self.modifications])

        return metadata_type + metadata_id + modifications

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load address alias-specific data data from catbuffer."""
        raise ValueError('Not implemented.')

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'metadataType', 'metadataId', 'modifications'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'metadataType': util.u8_to_dto(self.metadata_type),
            'metadataId': util.u64_to_dto(int(self.metadata_id)),
            'modifications': [MetadataModification.to_dto(x, network_type) for x in self.modifications],
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        metadata_type = MetadataType.create_from_dto(data['metadataType'], network_type)
        metadata_id = MosaicId(util.u64_from_dto(data['metadataId']))
        modifications = [MetadataModification.create_from_dto(x, network_type) for x in data['modifications']]

        self._set('metadata_type', metadata_type)
        self._set('metadata_id', metadata_id)
        self._set('modifications', modifications)


@register_transaction('MODIFY_MOSAIC_METADATA')
class ModifyMosaicMetadataInnerTransaction(InnerTransaction, ModifyMosaicMetadataTransaction):
    """Embedded mosaic metadata modification transaction."""

    __slots__ = ()



"""
    modify_multisig_account_transaction
    ===================================

    Modify multisig account transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .multisig_cosignatory_modification import MultisigCosignatoryModification
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = [
    'ModifyMultisigAccountTransaction',
    'ModifyMultisigAccountInnerTransaction',
]

MultisigCosignatoryModificationList = typing.Sequence[MultisigCosignatoryModification]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('MODIFY_MULTISIG_ACCOUNT')
class ModifyMultisigAccountTransaction(Transaction):
    """
    Modify multisig account transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param min_approval_delta: Minimum approval relative change.
    :param min_removal_delta: Minimum removal relative change.
    :param modifications: List of modifications.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    min_approval_delta: int
    min_removal_delta: int
    modifications: typing.Sequence[MultisigCosignatoryModification]

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        min_approval_delta: int,
        min_removal_delta: int,
        max_fee: int = 0,
        modifications: typing.Optional[MultisigCosignatoryModificationList] = None,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None,
    ) -> None:
        super().__init__(
            TransactionType.MODIFY_MULTISIG_ACCOUNT,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info,
        )
        self._set('min_approval_delta', min_approval_delta)
        self._set('min_removal_delta', min_removal_delta)
        self._set('modifications', modifications or [])

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        min_approval_delta: int,
        min_removal_delta: int,
        modifications: MultisigCosignatoryModificationList,
        network_type: NetworkType,
        max_fee: int = 0,
    ):
        """
        Create new transfer transaction.

        :param deadline: Deadline to include transaction.
        :param min_approval_delta: Minimum approval relative change.
        :param min_removal_delta: Minimum removal relative change.
        :param modifications: List of modifications.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.MODIFY_MULTISIG_ACCOUNT,
            deadline,
            min_approval_delta,
            min_removal_delta,
            max_fee,
            modifications
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        approval_size = util.I8_BYTES
        removal_size = util.I8_BYTES
        modification_size = MultisigCosignatoryModification.CATBUFFER_SIZE
        modifications_size = util.U8_BYTES + modification_size * len(self.modifications)
        return approval_size + removal_size + modifications_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export modify multisig account-specific data to catbuffer."""

        # int8 min_removal_delta
        # int8 min_approval_delta
        # uint8 modifications_count
        # CosignatoryModification { uint8 modification_type; uint8[32] key; }
        # CosignatoryModification[modifications_count] modifications
        removal = util.i8_to_catbuffer(self.min_removal_delta)
        approval = util.i8_to_catbuffer(self.min_approval_delta)
        modifications_count = util.u8_to_catbuffer(len(self.modifications))
        modifications = MultisigCosignatoryModification.sequence_to_catbuffer(
            self.modifications,
            network_type
        )

        return removal + approval + modifications_count + modifications

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load modify multisig account-specific data from catbuffer."""

        from_catbuffer = MultisigCosignatoryModification.sequence_from_catbuffer_pair
        # int8 min_removal_delta
        # int8 min_approval_delta
        # uint8 modifications_count
        # CosignatoryModification { uint8 modification_type; uint8[32] key; }
        # CosignatoryModification[modifications_count] modifications
        removal = util.i8_from_catbuffer(data[:1])
        approval = util.i8_from_catbuffer(data[1:2])
        modifications_count = util.u8_from_catbuffer(data[2:3])
        data = data[3:]
        modifications, data = from_catbuffer(data, modifications_count, network_type)

        self._set('min_removal_delta', removal)
        self._set('min_approval_delta', approval)
        self._set('modifications', modifications)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'minApprovalDelta', 'minRemovalDelta', 'modifications'}
        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'minApprovalDelta': util.u8_to_dto(self.min_approval_delta),
            'minRemovalDelta': util.u8_to_dto(self.min_removal_delta),
            'modifications': MultisigCosignatoryModification.sequence_to_dto(
                self.modifications,
                network_type
            ),
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        approval = util.i8_from_dto(data['minApprovalDelta'])
        removal = util.i8_from_dto(data['minRemovalDelta'])
        modifications = MultisigCosignatoryModification.sequence_from_dto(
            data.get('modifications', []),
            network_type
        )

        self._set('min_approval_delta', approval)
        self._set('min_removal_delta', removal)
        self._set('modifications', modifications)


@register_transaction('MODIFY_MULTISIG_ACCOUNT')
class ModifyMultisigAccountInnerTransaction(
    InnerTransaction,
    ModifyMultisigAccountTransaction
):
    """Embedded modify multisig account transaction."""

    __slots__ = ()



"""
    id_generator
    ============

    Generate integral identifiers from names.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import itertools
import re
import typing

from . import hashlib
from . import stdint

__all__ = [
    'generate_mosaic_id',
    'generate_sub_namespace_id',
    'generate_namespace_id',
]

FQN = re.compile(r'\A[a-z0-9][a-z0-9-_]*\Z')
NAMESPACE_MAX_DEPTH = 3
pack_u64 = stdint.u64_to_catbuffer
unpack_u32 = stdint.u32_iter_from_catbuffer
unpack_u64_dto = lambda x: list(itertools.islice(unpack_u32(x), 2))


def generate_mosaic_id(nonce: bytes, public_key: bytes) -> int:
    """
    Generate mosaic ID from nonce and public key.

    :param nonce: Mosaic nonce.
    :param owner: Account of mosaic owner.

    Example:
        .. code-block:: python

            >>> nonce = b'\x00\x00\x00\x00'
            >>> public_key = binascii.unhexlify(
            ...     '7D08373CFFE4154E129E04F0827E5F3D'
            ...     '6907587E348757B0F87D2F839BF88246'
            ... )
            >>> generate_mosaic_id(nonce, public_key)
            3456466875032780966
    """

    assert len(nonce) == 4
    assert len(public_key) == 32

    hasher = hashlib.sha3_256()
    hasher.update(nonce)
    hasher.update(public_key)
    result = unpack_u64_dto(hasher.digest())

    return stdint.u64_from_dto((result[0], result[1] & 0x7FFFFFFF))


def generate_sub_namespace_id(parent_id: int, name: str) -> int:
    """
    Generate namespace ID from child namespace name and parent ID.

    :param parent_id: Identifier for parent namespace.
    :param name: Child namespace name.

    Example:
        .. code-block:: python

           >>> generate_sub_namespace_id(0x84b3552d375ffa4b, "sample")
    """

    hasher = hashlib.sha3_256()
    hasher.update(pack_u64(parent_id))
    hasher.update(name.encode('ascii'))
    result = unpack_u64_dto(hasher.digest())
    child_id = stdint.u64_from_dto((result[0], result[1] | 0x80000000))

    return child_id


def generate_namespace_id(*names: str) -> typing.Sequence[int]:
    """
    Generate namespace ID from names.

    :param *names: Component names for a namespace.

    Example:
        .. code-block:: python

           >>> generate_namespace_id("sample")
           >>> generate_namespace_id("sample", "subpath")
           >>> generate_namespace_id("sample.subpath")
    """

    name = '.'.join(names)
    names = name.split('.')
    if name == '':
        return []
    elif len(names) > NAMESPACE_MAX_DEPTH:
        raise ValueError(f"Namespace '{name}' has too many parts.")
    elif not all(FQN.match(i) for i in names):
        raise ValueError(f"Invalid part name for namespace '{name}'.")

    ids = []
    parent_id = 0
    for name in names:
        parent_id = generate_sub_namespace_id(parent_id, name)
        ids.append(parent_id)

    return ids



"""
    transaction
    ===========

    Abstract base class for transaction objects.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .aggregate_transaction_info import AggregateTransactionInfo
from .deadline import Deadline
from .format import CatbufferFormat, DTOFormat
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType, OptionalNetworkType
from ... import util


TransactionInfoType = typing.Union[TransactionInfo, AggregateTransactionInfo]
TransactionBaseType = typing.TypeVar('TransactionBaseType', bound='TransactionBase')
TypeMap = typing.Mapping[TransactionType, typing.Type[TransactionBaseType]]


@util.inherit_doc
@util.dataclass(frozen=True)
class TransactionBase(util.Model):
    """
    Abstract, shared transaction base class.

    :param type: Transaction type.
    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param signature: Transaction signature (missing if embedded transaction).
    :param signer: Account of transaction creator.
    :param transaction_info: Transaction metadata.
    """

    # FIELDS

    type: TransactionType
    network_type: NetworkType
    version: TransactionVersion
    deadline: typing.Optional[Deadline]
    max_fee: typing.Optional[int]
    signature: typing.Optional[str]
    signer: typing.Optional[PublicAccount]
    transaction_info: typing.Optional[TransactionInfoType]

    # OVERRIDABLE CLASSVARS
    # The following classvars should be re-implemented for each
    # base transaction type.

    # Registry of transaction types to models. Allows users to
    # register custom models to customize serialization/deserialization
    # logic.
    TYPE_MAP: typing.ClassVar[TypeMap]

    # Data to simplify the serialization/deserialization of
    # transactions of a given type. Stores pre-computed
    # slices for individual fields, using dictionary lookups.
    # Using Python dictionaries is competitive with attribute lookup,
    # so it should be competitive with native performance.
    CATBUFFER: typing.ClassVar[CatbufferFormat]

    # Data to simplify the serialization/deserialization of
    # transactions of a given type. Provides high-level
    # routines, and also maps fields 1:1 to transaction DTO
    # keys.
    DTO: typing.ClassVar[DTOFormat]

    # INFO

    def is_unconfirmed(self) -> bool:
        """Is transaction pending to be included."""
        info = self.transaction_info
        return info is not None and info.is_unconfirmed()

    def is_confirmed(self) -> bool:
        """Is transaction already included."""
        info = self.transaction_info
        return info is not None and info.is_confirmed()

    def has_missing_signatures(self) -> bool:
        """Does the transaction have missing signatures."""
        info = self.transaction_info
        return info is not None and info.has_missing_signatures()

    def is_unannounced(self) -> bool:
        """Is transaction not known by the network."""
        return self.transaction_info is None

    # CATBUFFER

    @classmethod
    def catbuffer_size_shared(cls) -> int:
        """Get the shared size of the entity. Internal use only."""
        return cls.CATBUFFER.size_shared

    def catbuffer_size_specific(self) -> int:
        """Get the transaction-specific size of the entity. Internal use only."""
        raise util.AbstractMethodError

    def catbuffer_size(self) -> int:
        """Get the total size of the entity. Internal use only."""
        shared = self.catbuffer_size_shared()
        specific = self.catbuffer_size_specific()
        return shared + specific

    def to_catbuffer_shared(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export shared transaction data to catbuffer. Internal use only."""
        raise util.AbstractMethodError

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export transaction-specific data to catbuffer. Internal use only."""
        raise util.AbstractMethodError

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: util.FeeCalculationStrategy = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        # Check network type matches transaction.
        if network_type is not None and network_type != self.network_type:
            raise ValueError('Network type does not match transaction.')

        # Use fee calculation algorithm
        max_fee = util.calculate_fee(fee_strategy, self.max_fee, self.catbuffer_size())
        self._set('max_fee', max_fee)

        # Save shared and specific transaction data.
        shared = self.to_catbuffer_shared(self.network_type)
        specific = self.to_catbuffer_specific(self.network_type)

        return shared + specific

    def load_catbuffer_shared(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load shared transaction data from catbuffer. Internal use only."""
        raise util.AbstractMethodError

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load transaction-specific data from catbuffer. Internal use only."""
        raise util.AbstractMethodError

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[TransactionBaseType],
        data: typing.AnyStr,
        network_type: OptionalNetworkType = None,
    ) -> TransactionBaseType:
        return cls.create_from_catbuffer_pair(data, network_type)[0]

    @classmethod
    def create_from_catbuffer_pair(
        cls: typing.Type[TransactionBaseType],
        data: typing.AnyStr,
        network_type: OptionalNetworkType = None,
    ) -> typing.Tuple[TransactionBaseType, bytes]:
        """
        Deserialize object from catbuffer interchange format.

        If the cls is a direct subclass of `TransactionBase`,
        determine the transaction type and therefore the correct class.
        If the cls is not a direct subclass of `TransactionBase`
        (and therefore a finalized transaction model), use the
        class directly.

        :param data: Transaction data in catbuffer interchange format.
        :param network_type: Network type.
        """

        # Decode the data and check initial parameters.
        data = util.decode_hex(data, with_prefix=True)
        if len(data) < cls.catbuffer_size_shared():
            raise ValueError('Insufficient data to deserialize transaction.')

        # If we have a base class, find the correct derived class.
        if cls in TransactionBase.__subclasses__():
            cls = cls.CATBUFFER.find_transaction(cls.TYPE_MAP, data)
        inst = typing.cast(TransactionBaseType, cls.__new__(cls))

        # Load the network type and the total size and check transaction data.
        size = cls.CATBUFFER.load_size(data)
        nt = cls.CATBUFFER.load_network_type(data)
        if network_type is not None and network_type != nt:
            raise ValueError('Network type does not match transaction.')
        if len(data) < size:
            raise ValueError('Transaction data shorter than entity size.')

        # Load shared and specific transaction data.
        data = inst.load_catbuffer_shared(data, nt)
        data = inst.load_catbuffer_specific(data, nt)
        return inst, data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        """Validate transaction-specific fields in data-transfer object."""
        raise util.AbstractMethodError

    @classmethod
    def validate_dto_shared(cls, data: dict) -> bool:
        """Validate shared fields in data-transfer object."""
        raise util.AbstractMethodError

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'transaction'}
        all_keys = required_keys | {'meta'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, all_keys)
            and cls.validate_dto_shared(data['transaction'])
            and cls.validate_dto_specific(data['transaction'])
        )

    def to_dto_shared(
        self,
        network_type: NetworkType,
    ) -> dict:
        """Export shared transaction data to DTO. Internal use only."""
        raise util.AbstractMethodError

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        """Export transaction-specific data to DTO. Internal use only."""
        raise util.AbstractMethodError

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        # Check network type matches transaction.
        if network_type is not None and network_type != self.network_type:
            raise ValueError('Network type does not match transaction.')

        # Save shared and specific transaction data.
        dto = self.to_dto_shared(self.network_type)
        dto['transaction'].update(self.to_dto_specific(self.network_type))
        return dto

    def load_dto_shared(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        """Load shared transaction data from DTO. Internal use only."""
        raise util.AbstractMethodError

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        """Load transaction-specific data from DTO. Internal use only."""
        raise util.AbstractMethodError

    @classmethod
    def create_from_dto(
        cls: typing.Type[TransactionBaseType],
        data: dict,
        network_type: OptionalNetworkType = None,
    ) -> TransactionBaseType:
        """
        Deserialize object from DTO interchange format.

        If the cls is a direct subclass of `TransactionBase`,
        determine the transaction type and therefore the correct class.
        If the cls is not a direct subclass of `TransactionBase`
        (and therefore a finalized transaction model), use the
        class directly.

        :param data: Transaction data in DTO interchange format.
        :param network_type: Network type.
        """

        # If we have a base class, find the correct derived class.
        if cls in TransactionBase.__subclasses__():
            cls = cls.DTO.find_transaction(cls.TYPE_MAP, data)
        inst = typing.cast(TransactionBaseType, cls.__new__(cls))

        # Validate after we find the correct subclass.
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        # Load and check the network type.
        nt = cls.DTO.load_network_type(data)
        if network_type is not None and network_type != nt:
            raise ValueError('Network type does not match transaction.')

        # Load shared and specific transaction data.
        inst.load_dto_shared(data, nt)
        inst.load_dto_specific(data['transaction'], nt)
        return inst



from xpxchain import util
from tests import harness


def run_test(case, name, items):
    get = lambda name: getattr(util, name)
    for value, low, high, cat, dto in items:
        case.assertEqual(get(f'{name}_low')(value), low)
        case.assertEqual(get(f'{name}_high')(value), high)
        case.assertEqual(get(f'{name}_from_catbuffer')(cat), value)
        case.assertEqual(get(f'{name}_from_dto')(dto), value)
        case.assertEqual(get(f'{name}_to_catbuffer')(value), cat)
        case.assertEqual(get(f'{name}_to_dto')(value), dto)
        case.assertEqual(next(get(f'{name}_iter_from_catbuffer')(cat)), value)
        case.assertEqual(next(get(f'{name}_iter_from_dto')([dto])), value)
        case.assertEqual(next(get(f'{name}_iter_to_catbuffer')([value])), cat)
        case.assertEqual(next(get(f'{name}_iter_to_dto')([value])), dto)


class TestUnsignedInt(harness.TestCase):

    def test_exceptions(self):
        with self.assertRaises(OverflowError):
            util.u8_high(1 << 8)
        with self.assertRaises(OverflowError):
            util.u8_low(1 << 8)
        with self.assertRaises(OverflowError):
            util.u8_from_catbuffer(b'\x00\x00')
        with self.assertRaises(ValueError):
            next(util.u16_iter_from_catbuffer(b'\x00'))
        with self.assertRaises(OverflowError):
            util.u8_to_dto(1 << 8)
        with self.assertRaises(OverflowError):
            util.u8_from_dto(1 << 8)
        with self.assertRaises(OverflowError):
            util.u16_to_dto(1 << 16)
        with self.assertRaises(OverflowError):
            util.u16_from_dto(1 << 16)
        with self.assertRaises(OverflowError):
            util.u32_to_dto(1 << 32)
        with self.assertRaises(OverflowError):
            util.u32_from_dto(1 << 32)
        with self.assertRaises(OverflowError):
            util.u64_to_dto(1 << 64)
        with self.assertRaises(ArithmeticError):
            util.u64_from_dto([1 << 32, 0])
        with self.assertRaises(ArithmeticError):
            util.u64_from_dto([0, 1 << 32])
        with self.assertRaises(ArithmeticError):
            util.u64_from_dto([0])
        with self.assertRaises(ArithmeticError):
            util.u64_from_dto([0, 0, 0])
        with self.assertRaises(OverflowError):
            util.u128_to_dto(1 << 128)
        with self.assertRaises(ArithmeticError):
            util.u128_from_dto([[1 << 32, 0]])

    def test_u8(self):
        run_test(
            case=self,
            name='u8',
            items=[
                (0x00, 0x0, 0x0, b'\x00', 0x00),
                (0x01, 0x1, 0x0, b'\x01', 0x01),
                (0x0F, 0xF, 0x0, b'\x0F', 0x0F),
                (0x1F, 0xF, 0x1, b'\x1F', 0x1F),
            ],
        )

    def test_u16(self):
        run_test(
            case=self,
            name='u16',
            items=[
                (0x0000, 0x00, 0x00, b'\x00\x00', 0x0000),
                (0x0001, 0x01, 0x00, b'\x01\x00', 0x0001),
                (0x00FF, 0xFF, 0x00, b'\xFF\x00', 0x00FF),
                (0x01FF, 0xFF, 0x01, b'\xFF\x01', 0x01FF),
            ],
        )

    def test_u32(self):
        run_test(
            case=self,
            name='u32',
            items=[
                (0x00000000, 0x0000, 0x0000, b'\x00\x00\x00\x00', 0x00000000),
                (0x00000001, 0x0001, 0x0000, b'\x01\x00\x00\x00', 0x00000001),
                (0x0000FFFF, 0xFFFF, 0x0000, b'\xFF\xFF\x00\x00', 0x0000FFFF),
                (0x0001FFFF, 0xFFFF, 0x0001, b'\xFF\xFF\x01\x00', 0x0001FFFF),
            ],
        )

    def test_u64(self):
        run_test(
            case=self,
            name='u64',
            items=[
                (0x0000000000000000, 0x00000000, 0x00000000, b'\x00\x00\x00\x00\x00\x00\x00\x00', [0x00000000, 0x00000000]),
                (0x0000000000000001, 0x00000001, 0x00000000, b'\x01\x00\x00\x00\x00\x00\x00\x00', [0x00000001, 0x00000000]),
                (0x00000000FFFFFFFF, 0xFFFFFFFF, 0x00000000, b'\xFF\xFF\xFF\xFF\x00\x00\x00\x00', [0xFFFFFFFF, 0x00000000]),
                (0x00000001FFFFFFFF, 0xFFFFFFFF, 0x00000001, b'\xFF\xFF\xFF\xFF\x01\x00\x00\x00', [0xFFFFFFFF, 0x00000001]),
            ],
        )

    def test_u128(self):
        run_test(
            case=self,
            name='u128',
            items=[
                (0x00000000000000000000000000000000, 0x0000000000000000, 0x0000000000000000, b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', [[0x00000000, 0x00000000], [0x00000000, 0x00000000]]),
                (0x00000000000000000000000000000001, 0x0000000000000001, 0x0000000000000000, b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', [[0x00000001, 0x00000000], [0x00000000, 0x00000000]]),
                (0x0000000000000000FFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0x0000000000000000, b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x00\x00\x00\x00\x00\x00\x00\x00', [[0xFFFFFFFF, 0xFFFFFFFF], [0x00000000, 0x00000000]]),
                (0x0000000000000001FFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0x0000000000000001, b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x01\x00\x00\x00\x00\x00\x00\x00', [[0xFFFFFFFF, 0xFFFFFFFF], [0x00000001, 0x00000000]]),
            ],
        )



from __future__ import annotations
import enum
from .. import util

__all__ = [
    'FeeCalculationStrategy',
    'calculate_fee'
]


@util.inherit_doc
class FeeCalculationStrategy(util.U8Mixin, util.EnumMixin, enum.IntEnum):
    """ Fee calculation strategy """

    ZERO = 0
    LOW = 25
    MEDIUM = 250
    HIGH = 2500

    def description(self) -> str:
        return DESCRIPTION[self]


DESCRIPTION = {
    FeeCalculationStrategy.ZERO: "Zero fee calculation strategy",
    FeeCalculationStrategy.LOW: "Low fee calculation strategy",
    FeeCalculationStrategy.MEDIUM: "Medium fee calculation strategy",
    FeeCalculationStrategy.HIGH: "High fee calculation strategy",
}


def calculate_fee(strategy: FeeCalculationStrategy, max_fee: int, transaction_size: int) -> int:
    return max(max_fee, strategy * transaction_size)



"""
    alias
    =====

    Interface for aliases.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .alias_type import AliasType
from ..account.address import Address
from ..blockchain.network_type import OptionalNetworkType
from ..mosaic.mosaic_id import MosaicId
from ... import util

__all__ = ['Alias']

ValueType = typing.Optional[typing.Union[Address, MosaicId]]


def dto_to_kwds(
    data: dict,
    network_type: OptionalNetworkType = None,
) -> dict:
    """Convert data transfer object to keywords for initializer."""

    kwds: dict = {'type': AliasType.create_from_dto(data['type'], network_type)}
    if data['type'] == AliasType.NONE:
        kwds['value'] = None
    elif data['type'] == AliasType.ADDRESS:
        kwds['value'] = Address.create_from_encoded(data['address'])
    elif data['type'] == AliasType.MOSAIC_ID:
        kwds['value'] = MosaicId(util.u64_from_dto(data['mosaicId']))
    else:       # pragma: unreachable
        raise ValueError("Invalid data for Alias.create_from_dto.")

    return kwds


@util.inherit_doc
@util.dataclass(frozen=True, type=AliasType.NONE, value=None)
class Alias(util.DTO):
    """
    Alias for type definitions.

    :param value: Address or mosaic ID for alias.

    DTO Format:
        .. code-block:: yaml

            AliasDTO:
                type: integer
                mosaicId?: UInt64DTO
                # Hex(Address) (50-bytes)
                address?: string
    """

    type: AliasType
    value: ValueType

    def __post_init__(self):
        if self.type == AliasType.ADDRESS:
            same_type = isinstance(self.value, Address)
        elif self.type == AliasType.MOSAIC_ID:
            same_type = isinstance(self.value, MosaicId)
        elif self.type == AliasType.NONE:
            same_type = self.value is None
        else:       # pragma: unreachable
            same_type = False
        if not same_type:
            raise TypeError("Alias value and type do not match.")

    @property
    def address(self) -> Address:
        if self.type != AliasType.ADDRESS:
            raise ValueError("Alias does not store address.")
        return self.value

    @property
    def mosaic_id(self) -> MosaicId:
        if self.type != AliasType.MOSAIC_ID:
            raise ValueError("Alias does not store mosaic ID.")
        return self.value

    def __eq__(self, other) -> bool:
        if not isinstance(other, Alias):
            return False
        return self.astuple(recurse=False) == other.astuple(recurse=False)

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        data: dict = {'type': self.type.to_dto(network_type)}
        if self.type == AliasType.NONE:
            pass
        elif self.type == AliasType.ADDRESS:
            address = typing.cast(Address, self.value)
            data['address'] = util.hexlify(address.encoded)
        elif self.type == AliasType.MOSAIC_ID:
            mosaic_id = typing.cast(MosaicId, self.value)
            data['mosaicId'] = util.u64_to_dto(int(mosaic_id))
        else:       # pragma: unreachable
            raise ValueError("Invalid data for Alias.to_dto.")

        return data

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        return cls(**dto_to_kwds(data, network_type))



"""
    responses
    =========

    Mocked response data for the .

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import json
import os

DIR = os.path.dirname(os.path.realpath(__file__))
DATADIR = os.path.join(DIR, 'data')
ENDPOINT = os.environ.get('NIS2_ENDPOINT', '//localhost:3000')


def load_response(name):
    with open(os.path.join(DATADIR, name)) as f:
        data = json.load(f)
        data['content'] = data['content'].encode('utf8')
        return data


# ACCOUNT
ACCOUNT_INFO = {
    'Ok': load_response('account_info.json'),
}

ACCOUNTS_INFO = {
    'Ok': load_response('accounts_info.json'),
}

ACCOUNT_PROPERTIES = {
    'Ok': load_response('account_properties.json'),
}

ACCOUNTS_PROPERTIES = {
    'Ok': load_response('accounts_properties.json'),
}

ACCOUNT_NAMES = {
    'Ok': load_response('account_names.json'),
}

MULTISIG_INFO = {
    'Ok': load_response('multisig_info.json'),
}

MULTISIG_GRAPH_INFO = {
    'Ok': load_response('multisig_graph_info.json'),
}

# METADATA
ACCOUNT_METADATA = {
    'Ok': load_response('account_metadata.json'),
}

MOSAIC_METADATA = {
    'Ok': load_response('mosaic_metadata.json'),
}

NAMESPACE_METADATA = {
    'Ok': load_response('namespace_metadata.json'),
}

METADATA = {
    'Ok': load_response('metadata.json'),
}

METADATAS = {
    'Ok': load_response('metadatas.json'),
}

# CONFIG
CONFIG = {
    'Ok': load_response('config.json'),
}

UPGRADE = {
    'Ok': load_response('upgrade.json'),
}

# Node
INFO = {
    'Ok': load_response('node_info.json'),
}

TIME = {
    'Ok': load_response('node_time.json'),
}

# BLOCKCHAIN
BLOCK_INFO = {
    'Ok': load_response('block_info.json'),
}

BLOCKS_INFO = {
    'Ok': load_response('blocks_info.json'),
}

BLOCK_TRANSACTIONS = {
    'Ok': load_response('block_transactions.json'),
}

BLOCK_TRANSACTION_MERKLE = {
    'Ok': load_response('block_transaction_merkle.json'),
}

BLOCK_RECEIPTS = {
    'Ok': load_response('block_receipts.json'),
}

DIAGNOSTIC_SERVER = {
    'Ok': load_response('diagnostic_server.json'),
}

CHAIN_HEIGHT = {
    'Ok': load_response('chain_height.json'),
}

CHAIN_SCORE = {
    'Ok': load_response('chain_score.json'),
}

DIAGNOSTIC_BLOCKS_INFO = {
    'Ok': load_response('diagnostic_blocks_info.json'),
}

DIAGNOSTIC_STORAGE = {
    'Ok': load_response('diagnostic_storage.json'),
}

BLOCK_RECEIPTS = {
    'Ok': load_response('block_receipts.json'),
}

# MOSAIC
MOSAIC_INFO = {
    'Ok': load_response('mosaic_info.json'),
}

MOSAICS_INFO = {
    'Ok': load_response('mosaics_info.json'),
}

MOSAICS_NAMES = {
    'Ok': load_response('mosaic_names.json'),
}

# NAMESPACE
NAMESPACE = {
    'nem': load_response('namespace.json'),
}

NAMESPACES = {
    'nem': load_response('namespaces.json'),
}

NAMESPACE_NAMES = {
    'nem': load_response('namespace_names.json'),
}

# NETWORK
NETWORK_TYPE = {
    'MIJIN_TEST': load_response('mijin_test_network.json'),
}

# TRANSACTION
TRANSACTION = {
    'Ok': load_response('transaction.json'),
}

TRANSACTIONS = {
    'Ok': load_response('transactions.json'),
}

TRANSACTION_STATUS = {
    'Ok': load_response('transaction_status.json'),
}

TRANSACTION_STATUSES = {
    'Ok': load_response('transaction_statuses.json'),
}

ANNOUNCE = {
    'Ok': load_response('announce.json'),
}

ANNOUNCE_PARTIAL = {
    'Ok': load_response('announce_partial.json'),
}

ANNOUNCE_COSIGNATURE = {
    'Ok': load_response('announce_cosignature.json'),
}



"""
    network_config_transaction
    ====================

    Transfer transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .deadline import Deadline
from .inner_transaction import InnerTransaction
from .registry import register_transaction
from .transaction import Transaction
from .transaction_info import TransactionInfo
from .transaction_type import TransactionType
from .transaction_version import TransactionVersion
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = [
    'NetworkConfigTransaction',
    'NetworkConfigInnerTransaction',
]


@util.inherit_doc
@util.dataclass(frozen=True)
@register_transaction('NETWORK_CONFIG')
class NetworkConfigTransaction(Transaction):
    """
    Network configuration transaction.

    :param network_type: Network type.
    :param version: Transaction version.
    :param deadline: Deadline to include transaction.
    :param max_fee: Max fee for the transaction. Higher fees increase priority.
    :param applyHeightDelta: Blockchain height from which the configuration is applied.
    :param networkConfig: Network configuration.
    :param supportedEntityVersions: Supported entities.
    :param signature: (Optional) Transaction signature (missing if embedded transaction).
    :param signer: (Optional) Account of transaction creator.
    :param transaction_info: (Optional) Transaction metadata.
    """

    apply_height_delta: int
    network_config: str
    supported_entity_versions: str

    def __init__(
        self,
        network_type: NetworkType,
        version: TransactionVersion,
        deadline: Deadline,
        apply_height_delta: int,
        network_config: str,
        supported_entity_versions: str,
        max_fee: int = 0,
        signature: typing.Optional[str] = None,
        signer: typing.Optional[PublicAccount] = None,
        transaction_info: typing.Optional[TransactionInfo] = None
    ) -> None:
        super().__init__(
            TransactionType.TRANSFER,
            network_type,
            version,
            deadline,
            max_fee,
            signature,
            signer,
            transaction_info
        )
        self._set('apply_height_delta', apply_height_delta)
        self._set('network_config', network_config)
        self._set('supported_entity_versions', supported_entity_versions)

    @classmethod
    def create(
        cls,
        deadline: Deadline,
        network_type: NetworkType,
        apply_height_delta: int,
        network_config: str,
        supported_entity_versions: str,
        max_fee: int = 0,
    ):
        """
        Create new network configuration transaction.

        :param deadline: Deadline to include transaction.
        :param network_type: Network type.
        :param max_fee: (Optional) Max fee defined by sender.
        """
        return cls(
            network_type,
            TransactionVersion.TRANSFER,
            deadline,
            apply_height_delta,
            network_config,
            supported_entity_versions,
            max_fee,
        )

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        apply_height_delta_size = util.U64_BYTES
        extra_size = util.U8_BYTES + util.U16_BYTES
        network_config_size = util.U8_BYTES * len(self.network_config)
        supported_entity_versions_size = util.U8_BYTES * len(self.supported_entity_versions)

        return apply_height_delta_size + extra_size + network_config_size + supported_entity_versions_size

    def to_catbuffer_specific(
        self,
        network_type: NetworkType,
    ) -> bytes:
        """Export transfer-specific data to catbuffer."""

        # uint64_t apply_height_delta
        # uint16_t network_config_size
        # uint16_t supported_entity_versions_size
        # uint8_t[network_config_size] network_config
        # uint8_t[supported_entity_versions_size] supported_entity_versions
        apply_height_delta = util.u64_to_catbuffer(self.apply_height_delta)
        network_config_size = util.u16_to_catbuffer(len(self.network_config))
        supported_entity_versions_size = util.u16_to_catbuffer(len(self.supported_entity_versions))
        network_config = self.network_config.encode('utf-8')
        supported_entity_versions = self.supported_entity_versions.encode('utf-8')

        return apply_height_delta
        + network_config_size
        + supported_entity_versions_size
        + network_config
        + supported_entity_versions

    def load_catbuffer_specific(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        """Load transfer-specific data from catbuffer."""

        # uint64_t apply_height_delta
        # uint16_t network_config_size
        # uint16_t supported_entity_versions_size
        # uint8_t[network_config_size] network_config
        # uint8_t[supported_entity_versions_size] supported_entity_versions
        apply_height_delta = util.u64_from_catbuffer(data[:util.U64_BYTES])
        data = data[util.U64_BYTES:]
        network_config_size = util.u16_from_catbuffer(data[:util.U16_BYTES])
        data = data[util.U16_BYTES:]
        supported_entity_versions_size = util.u16_from_catbuffer(data[:util.U16_BYTES])
        data = data[util.U16_BYTES:]
        network_config = data[:network_config_size].decode('utf-8')
        data = data[:network_config_size]
        supported_entity_versions = data[:supported_entity_versions_size].decode('utf-8')
        data = data[:supported_entity_versions_size]

        self._set('apply_height_delta', apply_height_delta)
        self._set('network_config', network_config)
        self._set('supported_entity_versions', supported_entity_versions)

        return data

    # DTO

    @classmethod
    def validate_dto_specific(cls, data: dict) -> bool:
        required_keys = {'applyHeightDelta', 'networkConfig', 'supportedEntityVersions'}

        return cls.validate_dto_required(data, required_keys)

    def to_dto_specific(
        self,
        network_type: NetworkType,
    ) -> dict:
        return {
            'applyHeightDelta': util.u64_to_dto(self.apply_height_delta),
            'networkConfig': self.network_config,
            'supportedEntityVersions': self.supported_entity_versions,
        }

    def load_dto_specific(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        apply_height_delta = util.u64_from_dto(data['applyHeightDelta'])
        network_config = data['networkConfig']
        supported_entity_versions = data['supportedEntityVersions']

        self._set('apply_height_delta', apply_height_delta)
        self._set('network_config', network_config)
        self._set('supported_entity_versions', supported_entity_versions)


@register_transaction('NETWORK_CONFIG')
class NetworkConfigInnerTransaction(InnerTransaction, NetworkConfigTransaction):
    """Embedded network configuration transaction."""

    __slots__ = ()




from __future__ import annotations

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['CatapultConfig']


@util.inherit_doc
@util.dataclass(frozen=True)
class CatapultConfig(util.DTO):
    """
    Basic information describing an account.

    :param catapult_config: Catapult config.

    DTO Format:
        .. code-block:: yaml

            ConfigDTO:
                height: UInt64DTO
                networkConfig: string
                supportedEntityVersions: string

            CatapultConfigDTO:
                catapultConfig: ConfigDTO
    """

    height: int
    network_config: str
    supported_entity_versions: str

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'networkConfig'}
        required_l2 = {
            'height',
            'networkConfig',
            'supportedEntityVersions'
        }
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            # Level 2
            and cls.validate_dto_required(data['networkConfig'], required_l2)
            and cls.validate_dto_all(data['networkConfig'], required_l2)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        network_config = {
            'height': util.u64_to_dto(self.height),
            'networkConfig': self.network_config,
            'supportedEntityVersions': self.supported_entity_versions
        }

        return {
            'networkConfig': network_config,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        network_config = data['networkConfig']
        return cls(
            height=util.u64_from_dto(network_config.get('height', [0, 0])),
            network_config=network_config['networkConfig'],
            supported_entity_versions=network_config['supportedEntityVersions']
        )
